#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/wait.h>
int main(int argc, char *argv[]) {

    if(argc != 3) {
        printf("usage: ./addmx file1 file2\n");
        exit(EXIT_FAILURE);
    }

/* ------ create and read matrix ------ */

    FILE *fp1, *fp2;
    if((fp1 = fopen(argv[1],"r")) == NULL){
        perror("cannot open first file\n");
        exit(EXIT_FAILURE);
    }
    if((fp2 = fopen(argv[2],"r")) == NULL){
        perror("cannot open second file\n");
        exit(EXIT_FAILURE);
    }

    size_t str_size = 0;
    char* str = NULL;
    getline(&str, &str_size, fp1);
    int n = atoi(str);
    int m = atoi(str+=2);
    int matrix1[n][m], matrix2[n][m];
    int i = 0, j = 0;
    //first matrix
    while ( getline(&str, &str_size, fp1) > 0 ) {
        char* token = strtok(str," ");
        while( token != NULL ) {
            matrix1[i][j]=atoi(token);
            token=strtok(NULL," ");
            j++;
        }
        i++;
        j=0;
    }
    i=0;

    //ignore first line of second file
    getline(&str, &str_size, fp2);

    //second matrix
    while ( getline(&str, &str_size, fp2) > 0 ) {
        char* token = strtok(str," ");
        while( token != NULL ) {
            matrix2[i][j]=atoi(token);
            token=strtok(NULL," ");
            j++;
        }
        i++;
        j=0;
    }

    fclose(fp1);
    fclose(fp2);

/* ------ setup shared memory ------ */
    int *partials = mmap(NULL, (sizeof(matrix1))*3, PROT_READ|PROT_WRITE,
                         MAP_SHARED|MAP_ANONYMOUS, 0, 0);
    if(partials == MAP_FAILED){
        perror("mmap");
        exit(EXIT_FAILURE);
    }

/* ----- put matrixes inside shared memory ------ */
    int temp = 0;
    for(i = 0; i < n; i++) {
        for(j = 0; j < m; j++) {
            partials[j+temp] = matrix1[i][j];
            partials[j+n*m+temp] = matrix2[i][j];
        }
        temp += m;
    }

/* ----- make m processes and calculate in parallel ------ */
    temp = 0;
    for(i = 0; i < m; i++) {
        pid_t pid;
        if ((pid = fork()) < 0) {
            perror("fork");
            exit(EXIT_FAILURE);
        }
        if(pid == 0) {
            for (int j = 0; j < n; j++) {
                partials[j*m+(n*m)*2 + temp] = partials[j*m+temp] + partials[j*m+n*m+temp];
            }
            exit(EXIT_SUCCESS);
        }
        temp+=1;
    }

/* ----- wait for child processes ----- */
    for(i = 0; i < m; i++) {
        if ( waitpid(-1, NULL, 0) < 0) {
            perror("waitpid");
            exit(EXIT_FAILURE);
        }
    }

/* ----- print final matrix -----*/
    printf("%dx%d\n",n,m);
    temp = 0;
    for(i = 0; i < n; i++) {
        for (j = 0; j < m; j++)
            printf("%d ",  partials[j+temp+n*m*2]);
        printf("\n");
        temp += m;
    }

/* ------ release shared memory ------ */
    if (munmap(partials, sizeof(partials)) < 0) {
        perror("munmap");
        exit(EXIT_FAILURE);
    }
    exit(EXIT_SUCCESS);
}