#include <sys/wait.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define READ_END 0
#define WRITE_END 1

#define INPUTSIZE 2500
#define WORDSIZE 50

#define CYPHER_NAME "cypher.txt"

struct cypher{
    char first[WORDSIZE];
    char second[WORDSIZE];

    // If word == first then switch to second
    // If word == second then switch to first
};


int main(int argc, char* argv[]){

    //check if two or four arguments are present

    if (argc != 1 || stdin == NULL) {
        printf("usage: cypher < file_to_be_encripted [> destination_file]\n");
        return EXIT_FAILURE;
    }
    else if((fseek(stdin, 0, SEEK_END), ftell(stdin)) <= 0){
        printf("A file is needed as input!\n");
        printf("usage: cypher < file_to_be_encripted [> destination_file]\n");
        exit(EXIT_FAILURE);
    }
    

    int nbytes, gotoson[2], gotodad[2];
    pid_t pid;

    if (pipe(gotoson) < 0) {
        perror("pipe error");
        exit(EXIT_FAILURE);
    }

    if (pipe(gotodad) < 0) {
        perror("pipe error");
        exit(EXIT_FAILURE);
    }

    if ((pid = fork()) < 0) {
        perror("fork error");
        exit(EXIT_FAILURE);
    }

    else if (pid > 0) {
        /* PARENT PROCESS */
        close(gotoson[READ_END]); //The parent process shouldn't read, only write
        close(gotodad[WRITE_END]);

        /* passing input to child process */

        char buffer[INPUTSIZE];

        fseek(stdin, 0, SEEK_SET);

        char c = fgetc(stdin);

        memset(buffer, 0, INPUTSIZE);

        while (!feof(stdin)) {

            strncat(buffer, &c, 1);

            c = fgetc(stdin);
        }


        /* checks if something was written to the child process */

        if ((nbytes = write(gotoson[WRITE_END], buffer, strlen(buffer))) < 0) {
            fprintf(stderr, "Unable to write to pipe: %s\n", strerror(errno));
        }

        close(gotoson[WRITE_END]);


        /* wait for child and exit */

        if ( waitpid(pid, NULL, 0) < 0) {
            fprintf(stderr, "Cannot wait for child: %s\n", strerror(errno));
        }

        /* read the result from the pipe*/

        if ((nbytes = read(gotodad[READ_END], buffer, INPUTSIZE)) < 0 ) {
            fprintf(stderr, "Unable to read from pipe: %s\n", strerror(errno));
        }

        close(gotodad[READ_END]);
        write(STDOUT_FILENO, buffer, strlen(buffer));

        exit(EXIT_SUCCESS);
    }

    else {
        /* CHILD PROCESS */
        close(gotoson[WRITE_END]); //The child process shouldn't write, only read
        close(gotodad[READ_END]);

        char buffer[INPUTSIZE];

        if ((nbytes = read(gotoson[READ_END], buffer, INPUTSIZE)) < 0 ) {
            fprintf(stderr, "Unable to read from pipe: %s\n", strerror(errno));
        };

        close(gotoson[READ_END]);

        /* write message from parent */

        FILE *fcypher = fopen(CYPHER_NAME, "r");

        struct cypher list[50];
        char temp[WORDSIZE];
        int i=0, ccount = 0, lcount=0;
        char c;

        if(fcypher == NULL){
            printf("Error opening the cypher file\nMake sure you have a file named 'cypher.txt'\n");
            return EXIT_FAILURE;
        }
        else{
            memset(temp, 0, WORDSIZE);
            /* store words to be replaced and their replacements from the cypher file */
            c = fgetc(fcypher);
            while(!feof(fcypher)){
                
                if(c == ' '){
                    strncpy(list[i].first, temp, ccount-1);
                    memset(temp, 0, WORDSIZE);
                    ccount = 0;
                }
                else if(c == '\n'){
                    strncpy(list[i].second, temp, ccount-1);
                    lcount++;
                    memset(temp, 0, WORDSIZE);
                    i++;
                    ccount = 0;
                }
                else{
                    strncat(temp, &c, 1);
                    ccount++;
                }

                c = fgetc(fcypher);

            }

        }

        /* substitute words from file given to pipe*/
        char output[INPUTSIZE];
        memset(temp, 0, WORDSIZE);
        memset(output, 0, INPUTSIZE);
        int found;
        ccount = 0;

    

        for(int j=0; j<strlen(buffer); j++) {
            c = buffer[j];

            found = 0;

            if(c == ' ' || c == '.' || c == '?' || c == '!'|| c == ',' || c == ';' || c == '\n' || c == ':'){

                for(int k=0; k < lcount; k++){
                    if(strncmp(temp, list[k].first, strlen(list[k].first)) == 0){
                        strncat(output, list[k].second, strlen(list[k].second));
                        found = 1;
                        break;
                    }
                    else if(strncmp(temp, list[k].second, strlen(list[k].second)) == 0){
                        strncat(output, list[k].first, strlen(list[k].first));
                        found = 1;
                        break;
                    }
                }
                
                if(found == 0){
                    strncat(output, temp, strlen(temp));
                }

                strncat(output, &c, 1);
                memset(temp, 0, WORDSIZE);
                ccount = 0;
            }
            else{
                strncat(temp, &c, 1);
                ccount++;
            }       

        }

        write(gotodad[WRITE_END], output, strlen(output)-1);

        close(gotodad[WRITE_END]);

        /* exit gracefully */
        exit(EXIT_SUCCESS);
    }


    return EXIT_SUCCESS;
}